(* First, notation. *)
infix |>
fun x |> f = f x



datatype list
  = Nil
  | Cons of int * list

fun foldr g z Nil            = z
  | foldr g z (Cons (x, xs)) = g (x, foldr g z xs)

fun foldr (z, g) Nil            = z
  | foldr (z, g) (Cons (x, xs)) = g (x, foldr (z, g) xs)

fun foldr (nil', cons) Nil            = nil'
  | foldr (nil', cons) (Cons (x, xs)) = cons (x, foldr (nil', cons) xs)

fun foldr (nil', cons) =
  let
    fun replace Nil            = nil'
      | replace (Cons (x, xs)) = cons (x, replace xs)
  in
    replace
  end

datatype list
  = Nil  of unit
  | Cons of int * list

fun foldr (nil', cons) =
  let
    fun replace (Nil u)        = nil' u
      | replace (Cons (x, xs)) = cons (x, replace xs)
  in
    replace
  end


datatype tree
  = Leaf   of string
  | Branch of int * tree
  | Node   of tree * bool * tree

fun fold_tree (leaf, branch, node) =
  let
    fun replace (Leaf s)         = leaf s
      | replace (Branch (x, t))  = branch (x, replace t)
      | replace (Node (l, b, r)) = node (replace l, b, replace r)
  in
    replace
  end


signature FOLD =
  sig
    type t

    (* val fold : ???? -> t -> 'a *)
  (* val fold : (unit -> 'a) * (int * 'a -> 'a)                            -> t -> 'a *)
  (* val fold : (string -> 'a) * (int * 'a -> 'a) * ('a * bool * 'a -> 'a) -> t -> 'a *)
  (*
       (a^x) * (a^y) = a^(x + y)
       (x -> a) * (y -> a) = (x + y) -> a
  *)
  (* val fold : (unit + (int * 'a)                      -> 'a) -> t -> 'a *)
  (* val fold : (string + (int * 'a) + ('a * bool * 'a) -> 'a) -> t -> 'a *)

    type 'a template

    val fold : ('a template -> 'a) -> t -> 'a

  end


datatype 'a tree_template
  = T_Leaf of string
  | T_Branch of int * 'a
  | T_Node of 'a * bool * 'a


fun fold (f : 'a tree_template -> 'a) : tree -> 'a =
  let
    fun replace (Leaf s)           = f (T_Leaf s)
      | replace (Branch (x, t))    = f (T_Branch (x, replace t))
      | replace (Node (t1, b, t2)) = f (T_Node (replace t1, b, replace t2))  
  in
    replace
  end
(* 

(Node (?,true,?) : 'a tree_template)
(Node (Leaf "x",true,Leaf "u")) : ('a tree_template) tree_template


'a tree_template^n = trees that are <= n layers deep

tree = 'a tree_template^infinity

t = t tree_template

t = t tree_template tree_template

t = t tree_template tree_template tree_template 

 *)


datatype tree = WRAP of tree tree_template
(* WRAP : tree tree_template -> tree *)


val t1 : tree tree_template = T_Leaf "hi"
val t1' : tree = WRAP (T_Leaf "hi")

val t2 : tree = WRAP (T_Branch (3,WRAP (T_Leaf "hi")))


val tree =
  WRAP (T_Node (
    WRAP (T_Branch (
      42,
      WRAP (T_Leaf "hi")
    )),
    true,
    WRAP (T_Leaf "there")
  ))

val node   = WRAP o T_Node
val branch = WRAP o T_Branch
val leaf   = WRAP o T_Leaf


val tree =
  node (
    branch (
      42,
      leaf "hi"
    ),
    true,
    leaf "there"
  )

signature FOLD =
  sig
    type 'a template
    datatype t = WRAP of t template

    val wrap   : t template -> t
    val unwrap : t -> t template

    val fold : ('a template -> 'a) -> t -> 'a
  end


functor MkRecursive (
  type 'a template
  val map : ('a -> 'b) -> 'a template -> 'b template
) :> FOLD where type 'a template = 'a template =
struct
  type 'a template = 'a template
  datatype t = WRAP of t template
  
  val wrap = WRAP
  val unwrap = fn WRAP x => x

  fun fold (f : 'a template -> 'a) : t -> 'a = 
    let
      fun replace (WRAP templ) = f (map replace templ)
    in
      replace
    end
  
  fun fold f t = t |> unwrap |> map (fold f) |> f
end

structure TreeTemplate =
  struct
    datatype 'a template
      = Leaf   of string
      | Branch of int * 'a
      | Node   of 'a * bool * 'a

    val map = fn f =>
     fn Leaf x         => Leaf x
      | Branch (x, t)  => Branch (x, f t)
      | Node (l, b, r) => Node (f l, b, f r)
  end

structure Tree = MkRecursive (TreeTemplate)

structure NatTemplate =
  struct
    datatype 'a template
      = Zero
      | Succ of 'a

    val map = fn f =>
      fn Zero   => Zero
       | Succ x => Succ (f x)
  end
open NatTemplate

structure Nat = MkRecursive (NatTemplate)

val zero = Nat.wrap NatTemplate.Zero
val succ = Nat.wrap o NatTemplate.Succ

val ex1 =
  Nat.fold
    (fn Zero => 0 | Succ n => n + 1)
    (succ (succ (zero)))

val ex2 =
  Nat.fold
    (fn Zero => zero | Succ n => succ (succ n))
    (succ (succ (zero)))


structure ListTemplate =
  struct
    datatype 'a template
      = Nil
      | Cons of int * 'a

    val map = fn f =>
      fn Nil          => Nil
       | Cons (x, xs) => Cons (x, f xs)
  end

structure List = MkRecursive (ListTemplate)

val nil' = List.wrap ListTemplate.Nil
val cons = List.wrap o ListTemplate.Cons

val l = cons (3, cons (1, cons (7, nil')))

val 11 =
  let
    open ListTemplate
  in
    List.fold (fn Nil => 0 | Cons (x,xs) => x + xs) l
  end





functor MkCachedRecursive (
  type 'a template
  val map : ('a -> 'b) -> 'a template -> 'b template
  val compare : ('a * 'a -> order) -> 'a template * 'a template -> order
) =
  struct
    type 'a template = 'a template
    datatype t = WRAP of t template

    val wrap = WRAP
    fun unwrap (WRAP x) = x

    local
      fun compare' (WRAP template1, WRAP template2) = compare compare' (template1, template2)
    in
      val compare = compare'
    end

    structure S =
      SplayMapFn (
        type ord_key = t
        val compare = compare
      )

    fun fold f t =
      let
        val cache = ref S.empty
        fun aux t =
          case S.find (!cache, t) of
            NONE        =>
              let
                val result = t |> unwrap |> map aux |> f
              in
                cache := S.insert (!cache, t, result);
                result
              end
          | SOME result => result
      in
        aux t
      end
  end


structure ShrubTemplate =
  struct
    datatype 'a template
      = Leaf of int
      | Node of 'a * 'a

    val map = fn f =>
     fn Leaf x     => Leaf x
      | Node (l,r) => Node (f l, f r)

    fun compare f =
     fn (Leaf x, Leaf y) => Int.compare (x, y)
      | (Leaf _, Node _) => LESS
      | (Node _, Leaf _) => GREATER
      | (Node (l1,r1), Node (l2,r2)) => (
          case f (l1, l2) of
            EQUAL => f (r1, r2)
          | ord   => ord
        )

  end

structure Shrub = MkCachedRecursive (ShrubTemplate)

val leaf = Shrub.wrap o ShrubTemplate.Leaf
val node = Shrub.wrap o ShrubTemplate.Node

val op + = fn (x,y) => (print ("Adding " ^ Int.toString x ^ " and " ^ Int.toString y ^ "...\n"); x + y)

val sum =
  let
    open ShrubTemplate
  in
    Shrub.fold (fn Leaf x => x | Node (x,y) => x + y)
  end

fun fib 0 = leaf 0
  | fib 1 = leaf 1
  | fib n = node (fib (n - 1), fib (n - 2))

val fib30_tree = fib 30

(* datatype t = WRAP of unit -> t template *)


(* EXTRA CONTENT: Lazy Types *)

functor MkLazyRecursive (
  type 'a template
  val map : ('a -> 'b) -> 'a template -> 'b template
) =
  struct
    type 'a template = 'a template
    datatype t = WRAP of unit -> t template

    val wrap = WRAP
    fun unwrap (WRAP x) = x

    local
      fun poke x = x ()
    in
      fun fold f t = t |> unwrap |> poke |> map (fold f) |> f
    end
  end

(* We can re-use the exact same `ListTemplate` to get lazy lists - streams! *)
structure Stream = MkLazyRecursive (ListTemplate)

local
  open ListTemplate
  fun natsFrom n = Stream.wrap (fn () => ListTemplate.Cons (n, natsFrom (n + 1)))
in
  val nats = natsFrom 0

  fun take 0 stream = []
    | take n stream = (
        case (Stream.unwrap stream) () of
          Nil            => []
        | Cons (x, tail) => x :: take (n - 1) tail
      )
end
